/****************************************************************
 *                                                              *
 *                        I C A O   M a p  -                    *
 *                                                              *
 *             A n   A v i a t i o n   U t i l i t y            *
 *                                                              *
 *                                                              *
 *            Copyright (C) 1993, 1994  Martin Pauly            *
 *                                                              *
 *   This file may be freely distributed only if it includes    *
 *                the above copyright notice.                   *
 *                                                              *
 ****************************************************************/


/*    FVU file format of IGC    */

#include "fvu.h"
#include "gpssim.h"
#include "coord.h"

#include <stdio.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>


extern Widget toplevel;



static void exit_cb (Widget widget, XtPointer client_data, XtPointer call_data)
{
  fprintf (stderr, "exit\n");
}




static void process_line (char *line)
{
    int hours, minutes, seconds;        /* time of fix */
    int degrees;
    LOCATION fix;

    switch (line[0]) {
      case 'B':                 /* B Record (Fix) */
	if (line[24] == 'A') {      /* valid flag */
	    
	    /* time */
	    
	    hours   = (line[1]-'0') * 10 + line[2]-'0';
	    minutes = (line[3]-'0') * 10 + line[4]-'0';
	    seconds = (line[5]-'0') * 10 + line[6]-'0';
	    log_points[log_numpoints].time = (hours * 60 + minutes) * 60 + seconds;
	    
	    /* latitude */
	    
	    degrees = (line[7]-'0') * 10 + line[8]-'0';
	    minutes = (line[9]-'0') * 10000 + (line[10]-'0') * 1000 + (line[11]-'0') * 100 + 
	      (line[12]-'0') * 10 + line[13]-'0';
	    fix.latitude = 360000 * degrees + 6 * minutes;
	    if (line[14] == 'S') {
		fix.latitude *= -1;
	    }
	    
	    /* longitude */
	    
	    degrees = (line[15]-'0') * 100 + (line[16]-'0') * 10 + line[17]-'0';
	    minutes = (line[18]-'0') * 10000 + (line[19]-'0') * 1000 + (line[20]-'0') * 100 + 
	      (line[21]-'0') * 10 + line[22]-'0';
	    fix.longitude = 360000 * degrees + 6 * minutes;
	    if (line[23] == 'W') {
		fix.longitude *= -1;
	    }
	    
	    log_points[log_numpoints].loc.latitude = fix.latitude;
	    log_points[log_numpoints].loc.longitude = fix.longitude;

	    log_points[log_numpoints].status = 0;
	    
	    /* altitude */
	    
	    log_points[log_numpoints].alt = (line[25]-'0') * 10000 + (line[26]-'0') * 1000 +
	      (line[27]-'0') * 100 + (line[28]-'0') * 10 + line[29]-'0';
	    
	    /* calculate track and speed from last fix */
	    
	    if (log_numpoints == 0) {
		log_points[log_numpoints].track = 0;
		log_points[log_numpoints].speed = 0;
	    } else {
		int timediff = log_points[log_numpoints].time - log_points[log_numpoints-1].time;
		double dist;
		
		log_points[log_numpoints].track = truetrack (log_points[log_numpoints-1].loc, fix);
		dist = distance (log_points[log_numpoints-1].loc, fix) * 1852;   /* distance in m */
		if (timediff != 0) {
		    log_points[log_numpoints].speed = (int) ((dist / timediff) * 3.6);
		} else {
		    log_points[log_numpoints].speed = log_points[log_numpoints-1].speed;
		}
	    }
	    
	    log_numpoints++;
	}
	break;

      case 'H':      /* H Record  (header) */
	/* still need to test what this prefix means: */

	if ((line[1] == 'F') && (line[2] == 'D') && (line[3] == 'T') && (line[4] == 'E')) {
	    gps_date[0] = line[5];
	    gps_date[1] = line[6];
	    gps_date[2] = '.';
	    gps_date[3] = line[7];
	    gps_date[4] = line[8];
	    gps_date[5] = '.';
	    gps_date[6] = line[9];
	    gps_date[7] = line[10];
	    gps_date[8] = 0;	    
	}
	break;
    }
}



static void fvu_read_log (char *filename)
{
    char line[1024];
    
    FILE *in = fopen (filename, "r");
    
    if (in) {
	log_numpoints = 0;
	fgets (line, 1023, in);
	
	while (!feof(in)) {
	    process_line (line);
	    	    
	    fgets (line, 1023, in);
	}
	
	fclose (in);

	adjust_map_for_log();
	gps_evaluate();
    }
}





static void cancel_cb (Widget widget, XtPointer client_data, XtPointer call_data)
{
    XtUnmapWidget (XtParent (widget));
}



static void read_fvu_cb (Widget widget, XtPointer client_data, XtPointer call_data)
{
    char line[1024];
    XmString string;
    char *filename;
    
    /* get filename from dialog */
    
    XtVaGetValues (widget, 
		   XmNdirSpec, &string,
		   NULL);

    XmStringGetLtoR (string, XmFONTLIST_DEFAULT_TAG, &filename);
		
    fvu_read_log (filename);

    XtFree (filename);    
    XtUnmapWidget (XtParent (widget));
}






void fvu_load_file()
{
  Widget dialog, shell;
  Arg    args[10];
  int i = 0;

  XmString mask = XmStringCreateLocalized ("*.gps");

  shell = XtVaCreatePopupShell ("filesbshell", xmDialogShellWidgetClass, toplevel,
				XmNtitle,            "Load GPS File",
				XmNdeleteResponse,   XmDESTROY,
				XmNallowShellResize, True,
				NULL);
  
  dialog = XtVaCreateManagedWidget("FileBox", xmFileSelectionBoxWidgetClass, shell,
				   XmNdialogType, XmDIALOG_FILE_SELECTION, 
				   XmNpattern, mask,
				   NULL);

  XtSetSensitive (XmFileSelectionBoxGetChild (dialog, XmDIALOG_HELP_BUTTON), False);

  XmStringFree (mask);

  XtAddCallback (dialog, XmNokCallback, read_fvu_cb, NULL);
  XtAddCallback (dialog, XmNcancelCallback, cancel_cb, NULL);

  XtManageChild (shell); 
}




