/****************************************************************
 *                                                              *
 *                        I C A O   M a p  -                    *
 *                                                              *
 *             A n   A v i a t i o n   U t i l i t y            *
 *                                                              *
 *                                                              *
 *            Copyright (C) 1993, 1994  Martin Pauly            *
 *                                                              *
 *   This file may be freely distributed only if it includes    *
 *                the above copyright notice.                   *
 *                                                              *
 ****************************************************************/

/* gps simulator */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/List.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>
#include <Xm/DialogS.h>
#include <Xm/ToggleB.h>
#include <Xm/FileSB.h>

#include <X11/Xlib.h>

#include <math.h>

#include "objects.h"
#include "popup.h"
#include "busy.h"
#include "graph_primi.h"
#include "mapobjects.h"
#include "gpssim.h"
#include "icao.h"


#define FROM  0
#define VIA1  1
#define VIA2  2
#define VIA3  3
#define DEST  4

extern OBJECT *currentroute[];
extern Display *display;
extern int screen_num;
extern double mapWidthcm, mapHeightcm;
extern long map_scale;

extern GC gcplane;
extern Widget drawingarea;
extern int mapWidth, mapHeight;


extern Widget toplevel;

extern OBJECT **objectlist;       /* objects.c */
extern int db_objects;            /* number of objects */
extern Cursor pointer, waitpointer;

extern FILE *GPS;

static Widget trianglelist, shell, browser, kmfield, barodrawingarea = NULL;


LOGRECORD log_points[10000];    /* for 10+ hours... */
int log_numpoints = 0;
int log_visible = 0;
int log_rate_of_circling;

int log_climb_num = 0;
int log_climb_start[1000];
int log_climb_end[1000];

int log_take_off_index;
int log_landing_index;

static int cur_speed = 0, cur_track = 0, cur_alt = 0;
static int gpssim_active = 0;
extern LOCATION map_origin;
static LOCATION cur_pos;
char gps_date[20];


void gps_sim_sendstr(void)
{
  char line[256];
  int x, y, x1, y1, x2, y2;

  if (gpssim_active) {

    /* letzte Position loeschen */

    XSetPlaneMask (display, gcplane, plane_mask (1));
    XSetFunction (display, gcplane, GXclear);
    XSetFillStyle (display, gcplane, FillSolid);
    XFillRectangle (display, XtWindow (drawingarea), gcplane,
		    0, 0, mapWidth, mapHeight);

    /* Neue Position berechnen */

    cur_pos.latitude  += (int) (cos(cur_track*3.14/180) * (double) cur_speed / 111.0 * 360000.0 / 1600.0);
    cur_pos.longitude += (int) ((sin(cur_track*3.14/180) * (double) cur_speed / 111.0 * 360000.0 / 1600.0) 
      / cos (internal2rad (cur_pos.latitude)));
    

    /* Neue Position zeichnen */

    internal2window (cur_pos, &x, &y);

    x1 = (int) (x + 25.0 * sin((160.0+cur_track)*3.14/180.0));
    y1 = (int) (y - 25.0 * cos((160.0+cur_track)*3.14/180.0));
    x2 = (int) (x + 25.0 * sin((200.0+cur_track)*3.14/180.0));
    y2 = (int) (y - 25.0 * cos((200.0+cur_track)*3.14/180.0));

    gp_setcolor (BLACK);

    drawline (x, y, x1, y1);
    drawline (x, y, x2, y2);
    drawline (x1, y1, x2, y2);

    /* GPS Strings erzeugen: */

    sprintf (line, "$GPGGA,100000,%8.3f,N,%9.3f,E,1,04,2.8,%5.1f,M,0,M,*49",
	     ((cur_pos.latitude / 360000) * 100) + (((double) (cur_pos.latitude % 360000)) / 6000.0),
	     ((cur_pos.longitude / 360000) * 100) + (((double) (cur_pos.longitude % 360000)) / 6000.0),
	     (double) cur_alt);
    fprintf (GPS, "%s\n", line);

    sprintf (line, "$GPRMC,10000,A,%8.3f,N,%9.3f,E,%5.1f,%5.1f,220195,001.6,W*67",
	     ((cur_pos.latitude / 360000) * 100) + (((double) (cur_pos.latitude % 360000)) / 6000.0),
	     ((cur_pos.longitude / 360000) * 100) + (((double) (cur_pos.longitude % 360000)) / 6000.0),
	     (double) cur_speed, (double) cur_track);
    fprintf (GPS, "%s\n", line);
  }
}




/* close dialog */

static void cancelCB (Widget w, XtPointer client_data, XmPushButtonCallbackStruct *cbs)
{
  Widget parent;

  parent = XtParent (w);
  parent = XtParent (parent);

  XtPopdown (parent);
  gpssim_active = 0;
}





/* close form */

static void closeCB (Widget w, XtPointer client_data, XmPushButtonCallbackStruct *cbs)
{
  Widget shell = (Widget) client_data;

  XtUnmanageChild (shell);
  XtDestroyWidget (shell);
  gpssim_active = 0;
}



void new_speed (Widget w, XtPointer clientdata, XtPointer calldata)
{
  XmScaleCallbackStruct *cbs = (XmScaleCallbackStruct*) calldata;

  cur_speed = cbs->value;
}



void new_track (Widget w, XtPointer clientdata, XtPointer calldata)
{
  XmScaleCallbackStruct *cbs = (XmScaleCallbackStruct*) calldata;

  cur_track = cbs->value;
}


void new_alt (Widget w, XtPointer clientdata, XtPointer calldata)
{
  XmScaleCallbackStruct *cbs = (XmScaleCallbackStruct*) calldata;

  cur_alt = cbs->value;
}




/* suggest triangles, user interface */

void gps_sim_dialog (void)
{
  int index, i;
  Arg args[20];
  Widget start, stats, stats1, stats2, label, label1, bclose, speed, track, alt,
         b1000, bcancel, bsuggest, form, nameviewport, binfo, rowcol;
  XmString to, km, suggest, quit, tempstr;


  gpssim_active = 1;
  if (currentroute[FROM])
    cur_pos = currentroute[FROM]->location;
  else
    cur_pos = map_origin;


  /* popup shell */

  shell = XtVaCreateWidget ("gpssimdialog", xmDialogShellWidgetClass, toplevel,
			    XmNtitle,            "ICAO GPS-Simulator",
			    XmNdeleteResponse,   XmDESTROY,
			    XmNallowShellResize, True,
			    NULL);
  
  form = XtVaCreateWidget ("gpssimform", xmFormWidgetClass, shell, 
				  XmNautoUnmanage,     False,
				  NULL);



  tempstr = XmStringCreateSimple ("  Change Speed and Track using the scales:  ");
  label1 = XtVaCreateManagedWidget ("trianglelabel", xmLabelWidgetClass, form,
				    XmNlabelString,      tempstr,
				    XmNtopAttachment,    XmATTACH_FORM,
				    XmNleftAttachment,   XmATTACH_FORM,
				    XmNrightAttachment,  XmATTACH_FORM,
				    XmNalignment,        XmALIGNMENT_CENTER, 
				    NULL);
  XmStringFree (tempstr);  


  tempstr = XmStringCreateSimple ("Speed (km/h)");
  speed = XtVaCreateManagedWidget ("speed", xmScaleWidgetClass, form,
				   XmNtitleString,      tempstr,
				   XmNrightAttachment,  XmATTACH_FORM,
				   XmNrightOffset,      10,
				   XmNminimum,          0,
				   XmNmaximum,          1000,
				   XmNvalue,            0,
				   XmNshowValue,        TRUE,
				   XmNorientation,      XmHORIZONTAL,
				   XmNtopAttachment,    XmATTACH_WIDGET,
				   XmNtopWidget,        label1,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNleftOffset,       10,
				   NULL);
  XmStringFree (tempstr);  
  XtAddCallback (speed, XmNvalueChangedCallback, (XtCallbackProc) new_speed, shell);
  


  tempstr = XmStringCreateSimple ("Track");
  track = XtVaCreateManagedWidget ("track", xmScaleWidgetClass, form,
				   XmNtitleString,      tempstr,
				   XmNminimum,          0,
				   XmNmaximum,          359,
				   XmNvalue,            0,
				   XmNshowValue,        TRUE,
				   XmNorientation,      XmHORIZONTAL,
				   XmNtopAttachment,    XmATTACH_WIDGET,
				   XmNtopWidget,        speed,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNleftOffset,       10,
				   XmNrightAttachment,   XmATTACH_FORM,
				   XmNrightOffset,       10,
				   NULL);
  XmStringFree (tempstr);  
  XtAddCallback (track, XmNvalueChangedCallback, (XtCallbackProc) new_track, shell);


  tempstr = XmStringCreateSimple ("Altitude");
  alt = XtVaCreateManagedWidget ("alt", xmScaleWidgetClass, form,
				 XmNtitleString,      tempstr,
				 XmNminimum,          0,
				 XmNmaximum,          5000,
				 XmNvalue,            0,
				 XmNshowValue,        TRUE,
				 XmNorientation,      XmHORIZONTAL,
				 XmNtopAttachment,    XmATTACH_WIDGET,
				 XmNtopWidget,        track,
				 XmNleftAttachment,   XmATTACH_FORM,
				 XmNleftOffset,       10,
				 XmNrightAttachment,   XmATTACH_FORM,
				 XmNrightOffset,       10,
				 NULL);
  XmStringFree (tempstr);  
  XtAddCallback (alt, XmNvalueChangedCallback, (XtCallbackProc) new_alt, shell);


  /* Close Button */

  tempstr = XmStringCreateSimple ("Close");
  bclose = XtVaCreateManagedWidget ("gpsclose", xmPushButtonWidgetClass, form,
				    XmNtopAttachment,    XmATTACH_WIDGET,
				    XmNtopWidget,        alt,
				    XmNtopOffset,        5,
				    XmNleftAttachment,   XmATTACH_FORM,
				    XmNleftOffset,       10,
				    XmNlabelString,      tempstr,
				    NULL);
  XmStringFree (tempstr);

  XtAddCallback (bclose, XmNactivateCallback, (XtCallbackProc) closeCB, shell);




  /* display complete dialog */

  XtManageChild (form);  
  XtManageChild (shell);


}




/* barogram dialog code */

/* close form */

static void barocloseCB (Widget w, XtPointer client_data, XmPushButtonCallbackStruct *cbs)
{
    Widget shell = (Widget) client_data;
    
    XtUnmanageChild (shell);
    XtDestroyWidget (shell);
    
    barodrawingarea = NULL;
}


extern Window staticwindow;



void baro_draw (int width, int height)
{
  int top_marg = 20;
  int left_marg = 50;
  int right_marg = 20;
  int bot_marg = 30;
  int time_start, time_end, time;
  char *c;
  int max_alt, i;
  char tmp[60];
  int x, y, last_x, last_y;

  gp_setcolor (BLACK);
  gp_setlinestyle (1, SOLID);
  
  /* get times */

  time_start = (log_points[0].time / 3600) * 3600;
  time_end   = ((log_points[log_numpoints-1].time + 3599) / 3600) * 3600;

  /* write full hours scale */

  time = time_start;
  while (time < time_end) {
    int x, y;
    sprintf (tmp, "%2d:%2d", time / 3600, (time % 3600) / 60);
    if (tmp[3] == ' ') {
      tmp[3] = '0';
    }

    x = left_marg + (width-left_marg-right_marg) * (time-time_start)/(time_end-time_start);
    y = height - bot_marg;

    if (time % 1800 == 0) {
      gp_setcolor (BLACK);
    } else {
      gp_setcolor (LIGHTGRAY);
    }
    gp_drawline (x, top_marg, x, y);

    gp_setcolor (BLACK);
    if (time % 3600 == 0) {
      gp_drawline (x, y-1, x, y+2);
      gp_smalltext (x, y+15, tmp);
    }

    time += 600;   /* next hour */
  }


  /* find max altitude */

  max_alt = 0;
  for (i=0; i<log_numpoints; i++) {
    if (log_points[i].alt > max_alt) {
      max_alt = log_points[i].alt;
    }
  }

  switch (unit_altitude) {
    case mtr:
      sprintf (tmp, "max altitude: %d m", max_alt);
      break;
    case ft:
      sprintf (tmp, "max altitude: %d ft", (int) (max_alt * 3.28));
      break;
  }      
  gp_smalltext (width-right_marg-70, top_marg/2+5, tmp);

  /* date of flight */

  sprintf (tmp, "date of flight: %s", gps_date);
  gp_smalltext (left_marg+60, top_marg/2+5, tmp);

  /* minimum max alt: 1200 m */

  if (max_alt < 1200) {
    max_alt = 1200;
  }

  /* draw alt scale */


  switch (unit_altitude) {
    case mtr:
      i = 500;
      while (i < max_alt) {
	  int y = height-bot_marg - (height-bot_marg-top_marg) * i / max_alt;
	  
	  gp_setcolor (LIGHTGRAY);
	  gp_drawline (left_marg, y, width-right_marg, y);
	  
	  gp_setcolor (BLACK);
	  gp_drawline (left_marg-1, y, left_marg+2, y);
	  
	  sprintf (tmp, "%5d", i);
	  gp_smalltext (left_marg / 2, y+8, tmp);
	  
	  i += 500;
      }
      break;

    case ft:
      i = 1000;
      while (i < (max_alt*3.28)) {
	  int y = height-bot_marg - (height-bot_marg-top_marg) * i / (max_alt*3.28);
	  
	  gp_setcolor (LIGHTGRAY);
	  gp_drawline (left_marg, y, width-right_marg, y);
	  
	  gp_setcolor (BLACK);
	  gp_drawline (left_marg-1, y, left_marg+2, y);
	  
	  sprintf (tmp, "%5d", i);
	  gp_smalltext (left_marg / 2, y+8, tmp);
	  
	  i += 1000;
      }
      break;
  }


  /* draw frame */

  gp_setcolor (BLACK);
  gp_drawline (left_marg, top_marg, left_marg, height-bot_marg);
  gp_drawline (width-right_marg, top_marg, width-right_marg, height-bot_marg);
  gp_drawline (left_marg, height-bot_marg, width-right_marg, height-bot_marg);


  /* draw actual baro */

  last_x = left_marg + (width-left_marg-right_marg) * (log_points[0].time-time_start)/(time_end-time_start);
  last_y = height-bot_marg - (height-bot_marg-top_marg) * log_points[0].alt / max_alt;

  for (i=1; i<log_numpoints; i++) {
    x = left_marg + (width-left_marg-right_marg) * (log_points[i].time-time_start)/(time_end-time_start);
    y = height-bot_marg - (height-bot_marg-top_marg) * log_points[i].alt / max_alt;

    if ((x != last_x) || (y != last_y)) {
      gp_drawline (last_x, last_y, x, y);
      last_x = x;
      last_y = y;
    }
  }
}


void baro_expose_callback (Widget w, XtPointer data, XmDrawingAreaCallbackStruct *cbs)
{
    Window win, savewin;
    int width, height;
    XWindowAttributes winattr;
   
    if ((cbs->reason == XmCR_RESIZE) || (cbs && cbs->event)) {
	if ((cbs->reason == XmCR_RESIZE) || !(cbs->event->xexpose.count)) {
	    if (win = XtWindow (barodrawingarea)) {    
		XClearWindow (display, win);
		
		XGetWindowAttributes (XtDisplay(drawingarea), win, &winattr);
		
		width  = winattr.width;
		height = winattr.height;
		
		savewin = staticwindow;
		staticwindow = win;
		
		baro_draw (width, height);
		
		staticwindow = savewin;
	    }
	}
    }
}




void log_show_baro()
{
    Arg args[20];
    Widget shell, bclose, form;
    XmString tempstr;
    Pixel  bg_color;
    XSetWindowAttributes attributes;
    
    if (log_numpoints && !barodrawingarea) {
	/* popup shell */
	
	shell = XtVaCreatePopupShell ("barodialog", xmDialogShellWidgetClass, toplevel,
				      XmNtitle,            "ICAO Map Barogram",
				      XmNdeleteResponse,   XmDESTROY,
				      XmNallowShellResize, True,
				      NULL);
	
	form = XtVaCreateWidget ("baroform", xmFormWidgetClass, shell, 
				 XmNautoUnmanage,     False,
				 NULL);
	
	
	
	barodrawingarea = XtVaCreateManagedWidget ("drawingarea", xmDrawingAreaWidgetClass, 
						   form,
						   XmNtopAttachment,    XmATTACH_FORM,
						   XmNbottomAttachment, XmATTACH_FORM,
						   XmNbottomOffset,     35,
						   XmNleftAttachment,   XmATTACH_FORM,
						   XmNrightAttachment,  XmATTACH_FORM,
						   XmNborderWidth,      1,
						   XmNwidth,            640,
						   XmNheight,           200,
						   NULL);
	
	XtAddCallback (barodrawingarea, XmNexposeCallback, 
		       (XtCallbackProc) baro_expose_callback, NULL);
	XtAddCallback (barodrawingarea, XmNresizeCallback, 
		       (XtCallbackProc) baro_expose_callback, NULL);
	
	
	bg_color = WhitePixel (display, screen_num);
	XtVaSetValues (barodrawingarea,
		       XmNbackground,  bg_color,
		       NULL);
	
	
	tempstr = XmStringCreateSimple ("Close");
	bclose = XtVaCreateManagedWidget ("baroclose", xmPushButtonWidgetClass, form,
					  XmNtopAttachment,    XmATTACH_WIDGET,
					  XmNtopWidget,        barodrawingarea,
				    XmNtopOffset,        5,
					  XmNleftAttachment,   XmATTACH_FORM,
					  XmNleftOffset,       10,
					  XmNlabelString,      tempstr,
					  NULL);
	XmStringFree (tempstr);
	XtAddCallback (bclose, XmNactivateCallback, (XtCallbackProc) barocloseCB, shell);
	
	/* display complete dialog */
	
	XtManageChild (form);  
	XtManageChild (shell);
	
	
	/* enable backing store for map window, that speeds up quite a bit */
	
	attributes.backing_store = WhenMapped;  
	XChangeWindowAttributes (display, XtWindow (barodrawingarea), 
				 CWBackingStore, &attributes);
    }
}



static char *mstrtok(char *s)
{
  static char *save = NULL;
  char *ret;

  if (s)
    save = s;
  else
    s = save;

  if (save) {
    ret = save;
    while ((*s) && (*s != ',') && (*s != '\n') && (*s != '*'))
      s++;
    
    if ((*s == ',') || (*s == '*'))
      save = s+1;
    else
      save = NULL;

    *s = 0;
    return ret;
    
  }
  else
    return NULL;
}



static void gps_process (char *line)
{
    static long lx = -1, ly = -1;
    char *c;
    char tmp[256];
    int gps_quality, gps_numsat, gps_altitude, gps_valid;
    char gps_utc[200];
    LOCATION loc;
    int x, y;
    int track, speed;
    
    c = mstrtok (line);     
    
    if (!strcmp (c, "$GPGGA"))  { /* GPS Fix Data */
	
	c = mstrtok (NULL);         /* UTC */
	
	c = mstrtok (NULL);         /* Latitude */
	
	c = mstrtok (NULL);         /* N/S */
	
	c = mstrtok (NULL);         /* Longitude */
	
	c = mstrtok (NULL);         /* W/E */
	
	c = mstrtok (NULL);         /* GPS-Quality: 0=no GPS, 1=GPS, 2=DGPS */
	if (c)
	  gps_quality = atoi(c);
	
	c = mstrtok (NULL);         /* number of satellites used */
	if (c) {
	    gps_numsat = atoi(c);
	    if (gps_numsat < 3) {
		strcpy (tmp, "Keine GPS-Position");
		if (gps_numsat)
		  sprintf (tmp, "Keine GPS-Position, %d Satelliten verf\004gbar", gps_numsat);	  
	    } else {
		sprintf (tmp, "GPS: %d Satelliten verf\004gbar", gps_numsat);	  
	    }
	}
	
	c = mstrtok (NULL);         /* HDOP */
	
	c = mstrtok (NULL);         /* antenna height */
	if (c)  {
	    gps_altitude = atoi (c);
	}
	
	c = mstrtok (NULL);         /* M */
	
	c = mstrtok (NULL);         /* geoidal height height */
	
	c = mstrtok (NULL);         /* M */
	
	c = mstrtok (NULL);         /* age of DGPS Data */
	
	c = mstrtok (NULL);         /* DGPS ref station */
	
	c = mstrtok (NULL);         /* checksum */
    } else {   
	if (!strcmp (c, "$GPRMC"))  { /* Recommended Specific GPS Data */
	    c = mstrtok (NULL);	/* UTC */
	    if (c)
	      strcpy (gps_utc, c);
	
	    c = mstrtok (NULL);	/* status: A=valid, V=invalid */
	    if (c)  {
		if (*c == 'V')
		  gps_valid = 0;
		if (*c == 'A')
		  gps_valid = 1;
	    }
	
	    c = mstrtok (NULL);	/* latitude */
	
	    if (c) {
		loc.latitude = ((atoi(c) / 100) * 360000) + (int) (6000 * atof(c+2));  
	    }
	
	    c = mstrtok (NULL);	/* N/S */
	    if (c)
	      if (*c == 'S')
		loc.latitude *= -1;
	
	    c = mstrtok (NULL);	/* longitude */
	
	    if (c) {
		loc.longitude = ((atoi(c) / 100) * 360000) + (int) (6000 * atof(c+3)); 
	    } 
	
	    c = mstrtok (NULL);	/* W/E */
	    if (c)
	      if (*c == 'W')
		loc.longitude *= -1;
	
	    c = mstrtok (NULL);	/* speed over ground/knots */
	    speed = atoi(c) * 1.852;
	
	    c = mstrtok (NULL);	/* true course over ground */
	    track = atoi(c);
	
	    c = mstrtok (NULL);	/* date */
	    gps_date[0] = c[0];
	    gps_date[1] = c[1];
	    gps_date[2] = '.';
	    gps_date[3] = c[2];
	    gps_date[4] = c[3];
	    gps_date[5] = '.';
	    gps_date[6] = c[4];
	    gps_date[7] = c[5];
	    gps_date[8] = 0;
	
	    c = mstrtok (NULL);	/* magn. variation */
	
	    c = mstrtok (NULL);	/* E/W */
	
	    c = mstrtok (NULL);	/* checksum */
	
	    if (gps_valid) {
		log_points[log_numpoints].loc.latitude  = loc.latitude;
		log_points[log_numpoints].loc.longitude = loc.longitude;
	    
		c = gps_utc;      
		log_points[log_numpoints].time = ((10*(c[0]-'0')+c[1]-'0') * 60 
						  + (10*(c[2]-'0')+c[3]-'0')) * 60 
						    + (10*(c[4]-'0')+c[5]-'0');
	    
		log_points[log_numpoints].alt = gps_altitude;
		log_points[log_numpoints].speed = speed;
		log_points[log_numpoints].track = track;
		log_points[log_numpoints].status = 0;
	    
		log_numpoints++;
	    }
	}
    }
}



/* flatten altitude a little, useful for GPS altitude when pressure alt is not available */

void flatten_altitude(void)
{
    float flattened_alt;
    float alpha=0.005;
    int i, j;

    /* search take off and landing time */

    log_take_off_index = 0;

    for (i=0; (i<log_numpoints) && !log_take_off_index; i++) {
	int is_fast = 1;
	for (j=i; j<i+5; j++) {
	    if (log_points[j].speed < 60) {
		is_fast = 0;
	    }
	    if (is_fast) {
		log_take_off_index = i;
	    }
	}
    }

    log_landing_index  = log_numpoints -1;

    for (i=log_numpoints-1; (i>5) && (log_landing_index == log_numpoints-1); i--) {
	int is_fast = 1;
	for (j=i; j>i-5; j--) {
	    if (log_points[j].speed < 30) {
		is_fast = 0;
	    }
	    if (is_fast) {
		log_landing_index = i;
	    }
	}
    }

    flattened_alt = log_points[0].alt;

    for (i=1; i<log_numpoints; i++) {
	if (i >= log_take_off_index) {
	    alpha = 0.5;
	}

	if (i == log_landing_index) {
	    alpha = 0.005;
	}

	flattened_alt = (alpha * log_points[i].alt + (1 - alpha) * flattened_alt);
	log_points[i].alt = (int) flattened_alt;
    }
}



/* distinguish cruise/climb, statistics... */

void gps_evaluate(void)
{
    int i, j;
  
    /* search take off and landing time */

    log_take_off_index = 0;

    for (i=0; (i<log_numpoints) && !log_take_off_index; i++) {
	int is_fast = 1;
	for (j=i; j<i+5; j++) {
	    if (log_points[j].speed < 60) {
		is_fast = 0;
	    }
	    if (is_fast) {
		log_take_off_index = i;
	    }
	}
    }

    log_landing_index  = log_numpoints -1;

    for (i=log_numpoints-1; (i>5) && (log_landing_index == log_numpoints-1); i--) {
	int is_fast = 1;
	for (j=i; j>i-5; j--) {
	    if (log_points[j].speed < 30) {
		is_fast = 0;
	    }
	    if (is_fast) {
		log_landing_index = i;
	    }
	}
    }

    /* distinguish between cruise and climb */

    for (i=log_take_off_index; i<log_landing_index; i++) {
	int curr_time = log_points[i].time;
	int pastindex, j;
	int d_direction = 0;

	/* find index of nearest point that at least was 30 seconds ago */
    
	pastindex = i;
	while ((pastindex >= log_take_off_index) && (log_points[pastindex].time+30 >= curr_time)) {
	    pastindex--;
	}

	/* assume circling if less than 0.3 NM moved in last 30 seconds */
	
	if (distance (log_points[pastindex].loc, log_points[i].loc) < 0.3) {
	    for (j=pastindex; j<=i; j++) {
		log_points[j].status = 1;
	    }
	}
    }


    /* circling periods that are very close together are united */

    for (i=log_take_off_index+1; i<log_landing_index; i++) {
	/* look for a change from circle to cruise */

	if ((log_points[i].status == 0) && (log_points[i-1].status == 1)) {
	    int abort = 0;
	    for (j=i; (!abort) && (j<log_landing_index); j++) {
		if (log_points[j].time-30 > log_points[i].time) {
		    abort = 1;
		}
		if (log_points[j].status == 1) {
		    int k;
		    for (k=i; k<j; k++) {
			log_points[k].status = 1;
		    }
		    abort = 1;
		}
	    }
	}
    }  


    /* statistics */

    {
	int count = 0, circle_count = 0;
	for (i=log_take_off_index; i<log_landing_index; i++) {
	    count++;
	    if (log_points[i].status == 1) {
		circle_count++;
	    }
	}

	log_rate_of_circling = (int) 100 * circle_count / count;
    }

  

  
    /* find altitude gain and integrated climb of all thermals */

    log_climb_num = 0;

    for (i=log_take_off_index; i<log_landing_index; i++) {
	if (log_points[i].status == 1) {
	    for (j=i; (log_points[j].status == 1) && (j < log_landing_index); j++) {
		/* do nothing, just increase j */
	    }

	    if (i < log_landing_index) {
		log_climb_start[log_climb_num] = i;
		log_climb_end[log_climb_num] = j;
		log_climb_num++;
	    }

	    i = j;		/* look for next circling */
	}
    }

}




static void cancel_cb (Widget widget, XtPointer client_data, XtPointer call_data)
{
    XtUnmapWidget (XtParent (widget));
}


static void read_gps_cb (Widget widget, XtPointer client_data, XtPointer call_data)
{
    char line[1024];
    XmString string;
    char *filename;
    
    /* get filename from dialog */
    
    XtVaGetValues (widget, 
		   XmNdirSpec, &string,
		   NULL);

    XmStringGetLtoR (string, XmFONTLIST_DEFAULT_TAG, &filename);
		
    gps_read_log (filename);

    XtFree (filename);    
    XtUnmapWidget (XtParent (widget));
}


void gps_read_log_dialog (void)
{
  Widget dialog, shell;
  Arg    args[10];
  int i = 0;

  XmString mask = XmStringCreateLocalized ("*.nmea");

  shell = XtVaCreatePopupShell ("filesbshell", xmDialogShellWidgetClass, toplevel,
				XmNtitle,            "Load NMEA File",
				XmNdeleteResponse,   XmDESTROY,
				XmNallowShellResize, True,
				NULL);
  
  dialog = XtVaCreateManagedWidget("FileBox", xmFileSelectionBoxWidgetClass, shell,
				   XmNdialogType, XmDIALOG_FILE_SELECTION, 
				   XmNpattern, mask,
				   NULL);

  XtSetSensitive (XmFileSelectionBoxGetChild (dialog, XmDIALOG_HELP_BUTTON), False);

  XmStringFree (mask);

  XtAddCallback (dialog, XmNokCallback, read_gps_cb, NULL);
  XtAddCallback (dialog, XmNcancelCallback, cancel_cb, NULL);

  XtManageChild (shell); 
}



/* make map display current log */

void adjust_map_for_log()
{
    if (log_numpoints > 10) {         /* don't do anything for pseudo files */
	int i;
	float heightNM, widthNM;
	LOCATION topleft, botright, a, b;
	double scalehor, scalever;

	topleft  = log_points[0].loc;
	botright = log_points[0].loc;

	for (i=1; i<log_numpoints; i++) {
	    if (log_points[i].loc.longitude < topleft.longitude) {
		topleft.longitude = log_points[i].loc.longitude;
	    }

	    if (log_points[i].loc.longitude > botright.longitude) {
		botright.longitude = log_points[i].loc.longitude;
	    }

	    if (log_points[i].loc.latitude < botright.latitude) {
		botright.latitude = log_points[i].loc.latitude;
	    }

	    if (log_points[i].loc.latitude > topleft.latitude) {
		topleft.latitude = log_points[i].loc.latitude;
	    }
	}

	a.longitude = b.longitude = topleft.longitude;
	a.latitude = topleft.latitude;
	b.latitude = botright.latitude;
	heightNM = distance (a, b);
	
	b.longitude = botright.longitude;
	b.latitude  = topleft.latitude;
	widthNM = distance (a, b);
	
	scalehor = widthNM * 185200 / mapWidthcm;
	scalever = heightNM * 185200 / mapHeightcm;
	
	map_scale = (long) ((scalehor > scalever) ? scalehor : scalever);
	map_scale = (long) (map_scale * 1.1);

	if (map_scale > 5000000) {
	    map_scale = 5000000;
	}
	
	/* last but not least: choose map origin so that everything will fit */
	
	map_origin.longitude = (long) ((topleft.longitude + botright.longitude) / 2);
	map_origin.latitude  = (long) ((topleft.latitude  + botright.latitude)  / 2);

	map_initcoord();
	log_visible = 1;
	
	XClearArea (display, XtWindow (drawingarea), 0, 0, 0, 0, True);		  	
    }
}



void gps_read_log (char *filename)
{
    char line[256];
    FILE *f = fopen (filename, "r");
    
    if (f) {
	log_numpoints = 0;
	fgets (line, 255, f);
	
	while (!feof(f)) {
	    gps_process (line);
	    
	    fgets (line, 255, f);
	}
	
	close (f);
	
	adjust_map_for_log();
	flatten_altitude();      /* no pressure altitude available */
	gps_evaluate();
    }
}



void gps_draw_log()
{
  int x, y, lx, ly, i, last_status = 0;
  int minx, maxx, miny, maxy;
  int xcent, ycent, alt1, alt2, time;
  char str[256];

  if (log_numpoints && log_visible) {
    internal2window (log_points[0].loc, &lx, &ly);
    minx = maxx = lx;
    miny = maxy = ly;

    gp_setcolor (BLACK);
    gp_setlinestyle (1, SOLID);

    for (i=log_take_off_index; i<log_landing_index; i++) {
      internal2window (log_points[i].loc, &x, &y);

      log_points[i].x = x;
      log_points[i].y = y;

      if (x > maxx) maxx = x;
      if (y > maxy) maxy = y;
      if (x < minx) minx = x;
      if (y < miny) miny = y;

      if ((lx != x) || (ly != y)) {
	if (log_points[i].status != last_status) {
	  if (log_points[i].status == 0) {
	    gp_setcolor (BLACK);
	  } else {
	    gp_setcolor (LIGHTGRAY);
	  }
	  last_status = log_points[i].status;
	}
	gp_drawline (lx, ly, x, y);
      }
      lx = x; ly = y;
    }

    /* middle of route */

    xcent = (minx + maxx) / 2;
    ycent = (miny + maxy) / 2;

    /* draw climb info */

    gp_setcolor (BLACK);
    for (i=0; i<log_climb_num; i++) {
      /* determine which end of current circling to use for the label */

      internal2window (log_points[log_climb_start[i]].loc, &x, &y);
      internal2window (log_points[log_climb_end[i]].loc, &lx, &ly);

      if (abs(ly-ycent) > abs(y-ycent)) {
	y = ly;
      }
      x = (x + lx) / 2;
      
      if (y > ycent) {     /* show info under circling */
	y += 20;
      } else {
	y -= 5;
      }
      
      alt1 = (log_points[log_climb_start[i]].alt + log_points[log_climb_start[i]+1].alt +
	      log_points[log_climb_start[i]+2].alt) / 3;
      alt2 = (log_points[log_climb_end[i]].alt + log_points[log_climb_end[i]-1].alt +
	      log_points[log_climb_end[i]-2].alt) / 3;
      time = log_points[log_climb_end[i]].time - log_points[log_climb_start[i]].time;
      
      switch (unit_altitude) {
	case mtr:      
	  sprintf (str, "%dm/%3.1fm/s", alt2-alt1, (float) ((float) (alt2-alt1) / time));
	  break;

	case ft:
	  sprintf (str, "%dft/%3.1fm/s", ((int) ((alt2-alt1)*3.28)), (float) ((float) (alt2-alt1) / time));
	  break;
      }

      gp_smalltext (x, y, str);
      
    }

  }
}



void baro_mark_time (int time)
{
    int x;
    int width, height;
    Window win, savewin;
    XWindowAttributes winattr;

    if (barodrawingarea) {
	if (win = XtWindow (barodrawingarea)) {    
	    int top_marg = 20;
	    int left_marg = 50;
	    int right_marg = 20;
	    int bot_marg = 30;
	    int time_start, time_end;
	    
	    XGetWindowAttributes (XtDisplay(drawingarea), win, &winattr);
	    width  = winattr.width;
	    height = winattr.height;
	    
	    savewin = staticwindow;
	    staticwindow = win;

	    /* alte Markierung loeschen */

	    gp_setcolor (WHITE);
	    gp_setlinestyle (1, SOLID);
	    gp_drawfilledrect (0, height-10, width, height-2);
	    
	    /* calc x position of marker */
	    
	    time_start = (log_points[0].time / 3600) * 3600;
	    time_end   = ((log_points[log_numpoints-1].time + 3599) / 3600) * 3600;	    
	    x = left_marg + (width-left_marg-right_marg) * (time-time_start)/(time_end-time_start);
	    
	    /* neue Markierung machen */
	    
	    if (time != 0) {
		gp_setcolor (BLACK);
		
		gp_drawline (x, height-2, x, height-10);
		
		gp_drawline (x-1, height-2, x-1, height-8);
		gp_drawline (x+1, height-2, x+1, height-8);
		
		gp_drawline (x-2, height-2, x-2, height-6);
		gp_drawline (x+2, height-2, x+2, height-6);
		
		gp_drawline (x-3, height-2, x-3, height-4);
		gp_drawline (x+3, height-2, x+3, height-4);
	    }

	    staticwindow = savewin;	    
	}
    }
}
