/****************************************************************
 *                                                              *
 *                        I C A O   M a p  -                    *
 *                                                              *
 *             A n   A v i a t i o n   U t i l i t y            *
 *                                                              *
 *                                                              *
 *            Copyright (C) 1993, 1994  Martin Pauly            *
 *                                                              *
 *   This file may be freely distributed only if it includes    *
 *                the above copyright notice.                   *
 *                                                              *
 ****************************************************************/



#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "setup.h"
#include "icao.h"

#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/DialogS.h>
#include <Xm/ToggleB.h>
#include <Xm/RowColumn.h>

#include "icao.h"
#include "coord.h"

extern long map_scale;
extern LOCATION map_origin;

extern Widget toplevel;

extern int visobj[];       /* info about visible objects */
extern int numvis;


enum entrytype { enumtype, inttype, stringtype , separator };


typedef struct {
    char *title;		/* title of form */
    char **option_strings;	/* strings to be displayed as options (0..n-1) */
    int  *value;		/* pointer to corresponding value (0..n-1) */
    int  num_options;		/* number of options */
    int  changed_selection;	/* selection was changed by user to this value */
} enum_entry;


typedef struct {
    char *title;		/* title of form */
    char *value;		/* pointer to string */
} string_entry;


typedef struct {
    char *title;		/* title of form */
    int  *value;		/* pointer to value */
} int_entry;


/* arrays with pointers to XXX_entry instances and type information */

static void *entries[50];
static enum entrytype types[50];
static int  num_entries;



#define NEWSTRING() num_entries++; entries[num_entries] = malloc (sizeof (string_entry)); types[num_entries] = stringtype
#define NEWENUM()   num_entries++; entries[num_entries] = malloc (sizeof (enum_entry));   types[num_entries] = enumtype
#define NEWINT()    num_entries++; entries[num_entries] = malloc (sizeof (int_entry));    types[num_entries] = inttype
#define NEWSEP()    num_entries++; entries[num_entries] = malloc (sizeof (int_entry));    types[num_entries] = separator

#define STRING ((string_entry *) entries[num_entries])
#define ENUM   ((enum_entry *) entries[num_entries])
#define INT    ((int_entry *) entries[num_entries])

static char *dist_strings[] = { "km   ", "NM   " };
static char *speed_strings[] = { "km/h ", "knots", "m/s  " };
static char *altitude_strings[] = { "m    ", "ft   " };
static char *projection_strings[] = { "Lambert", "Mercator" };
static char *projection_details_strings[] = { "manual  ", "auto" };
static char *dist_calc_strings[] = { "Pauly", "FAI" };

void setup_init (void)
{
    num_entries = -1;
  
    NEWENUM();
    ENUM->title = "Distance Unit:";
    ENUM->value = (int *) &unit_distance;
    ENUM->option_strings = dist_strings;
    ENUM->num_options = 2;

    NEWENUM();
    ENUM->title = "Speed Unit:";
    ENUM->value = (int *) &unit_speed;
    ENUM->option_strings = speed_strings;
    ENUM->num_options = 3;

    NEWENUM();
    ENUM->title = "Altitude Unit:";
    ENUM->value = (int *) &unit_altitude;
    ENUM->option_strings = altitude_strings;
    ENUM->num_options = 2;

    NEWSEP();

    NEWENUM();
    ENUM->title = "Projection Type:";
    ENUM->value = (int *) &map_projection;
    ENUM->option_strings = projection_strings;
    ENUM->num_options = 2;

    NEWENUM();
    ENUM->title = "Projection Details:";
    ENUM->value = (int *) &autoparallels;
    ENUM->option_strings = projection_details_strings;
    ENUM->num_options = 2;

    NEWSEP();

    NEWENUM();
    ENUM->title = "Distance Calculation:";
    ENUM->value = (int *) &config_distance_func;
    ENUM->option_strings = dist_calc_strings;
    ENUM->num_options = 2;


    num_entries++;
}



static void okCB (Widget w, XtPointer client_data, XmPushButtonCallbackStruct *cbs)
{
    int changed = 0, i;

    Widget parent = (Widget) client_data;
    XtUnmapWidget (parent);
  
    for (i=0; i<num_entries; i++) {
	switch (types[i]) {
	  case enumtype:
#define entry ((enum_entry *) entries[i])

	    if (*(entry->value) != entry->changed_selection) {
		changed = 1;
		*(entry->value) = entry->changed_selection;
	    }

#undef entry      
	    break;
	}
    }

    if (changed) {
	map_initcoord();
	XClearArea (display, XtWindow (drawingarea), 0, 0, 0, 0, True);	/* redraw */
    }

}




static void cancelCB (Widget w, XtPointer client_data, XmPushButtonCallbackStruct *cbs)
{
    Widget parent = (Widget) client_data;
    XtUnmapWidget (parent);
}




void buttonhitCB (Widget w, XtPointer client_data, XmToggleButtonCallbackStruct *cbs)
{
    if (cbs->set == True) {
	int i, j;
	Widget button = (Widget) client_data, rowcol;
	
	XtVaGetValues (button, XmNuserData, &j, NULL);
	rowcol = XtParent (button);
	XtVaGetValues (rowcol, XmNuserData, &i, NULL);
	
	((enum_entry *) entries[i])->changed_selection = j;
    }
}



void setup_dialog (void)
{
    Widget     shell, form, okbutton, cancelbutton, lastform = NULL, group, label, rowcol;
    XmString   string;
    int i, j;


    shell = XtVaCreatePopupShell ("setupdialog", xmDialogShellWidgetClass, toplevel,
				  XmNtitle,            "Options",
				  XmNdeleteResponse,   XmDESTROY,
				  XmNmappedWhenManaged, False,
				  XmNallowShellResize, True,
				  NULL);
  
    form = XtVaCreateManagedWidget ("setupform", xmFormWidgetClass, shell, 
				    NULL);
  
  
    /* generate entries for setup items */

    for (i=0; i<num_entries; i++) {
	switch (types[i]) {



	  case enumtype: 
#define entry ((enum_entry *) entries[i])

	    group = XtVaCreateManagedWidget ("form", xmFormWidgetClass, form, 
					     XmNtopAttachment,    lastform ? XmATTACH_WIDGET : XmATTACH_FORM,
					     XmNtopWidget,        lastform,
					     XmNtopOffset,        10,				       
					     NULL);      

	    string = XmStringCreateSimple (entry->title);
	    label = XtVaCreateManagedWidget ("label", xmLabelWidgetClass, group,
					     XmNlabelString,      string,     
					     XmNleftAttachment,   XmATTACH_FORM,
					     XmNleftOffset,       10,				       
					     XmNalignment,        XmALIGNMENT_BEGINNING,
					     NULL);
	    XmStringFree (string);

	    /* generate radio buttons for choices */

	    rowcol = XtVaCreateManagedWidget ("projmanager", xmRowColumnWidgetClass, form,
					      XmNradioBehavior,    True,
					      XmNorientation,      XmHORIZONTAL,
					      XmNnumColumns,       1,
					      XmNtopAttachment,    lastform ? XmATTACH_WIDGET : XmATTACH_FORM,
					      XmNtopWidget,        lastform,
					      XmNtopOffset,        5,
					      XmNleftAttachment,   XmATTACH_FORM,
					      XmNleftOffset,       150,
					      XmNrightAttachment,  XmATTACH_FORM,
					      XmNrightOffset,      10,
					      XmNuserData,         i, /* index in entries array */
					      NULL);

	    entry->changed_selection = *(entry->value);	/* default: user didn't change anything */

	    for (j=0; j<entry->num_options; j++) {
		Widget button;
		String label;
	
		string = XmStringCreateSimple (entry->option_strings[j]);

		button = XtVaCreateManagedWidget ("radiosetup", xmToggleButtonWidgetClass, 
						  rowcol,
						  XmNlabelString, string,
						  XmNset,         (*(entry->value) == j),
						  XmNuserData,    j, /* index of choice for element */
						  NULL);

		XtAddCallback (button, XmNvalueChangedCallback, (XtCallbackProc) buttonhitCB,
			       (XtPointer) button);

		XmStringFree (string);

	    }


	    lastform = group;
	    break;

#undef entry      



	  case inttype:
	    break;




	  case stringtype:
	    break;





	  case separator:
	    group = XtVaCreateManagedWidget ("form", xmFormWidgetClass, form, 
					     XmNtopAttachment,    lastform ? XmATTACH_WIDGET : XmATTACH_FORM,
					     XmNtopWidget,        lastform,
					     XmNtopOffset,        10,				       
					     NULL);      
	    lastform = group;

	    string = XmStringCreateSimple ("-------------------------------------------------------");
	    label = XtVaCreateManagedWidget ("label", xmLabelWidgetClass, group,
					     XmNlabelString,      string,     
					     XmNalignment,        XmALIGNMENT_BEGINNING,
					     NULL);
	    XmStringFree (string);

	    break;
	}
    }
    
    
    
    /* add OK and Cancel button */
    
    string = XmStringCreateSimple (" OK ");
    okbutton = XtVaCreateManagedWidget ("setupok", xmPushButtonWidgetClass, form,
					XmNlabelString,      string,
					XmNtopAttachment,    lastform ? XmATTACH_WIDGET : XmATTACH_FORM,
					XmNtopWidget,        lastform,
					XmNtopOffset,        10,
					XmNbottomAttachment, XmATTACH_FORM,
					XmNbottomOffset,     10,
					XmNleftAttachment,   XmATTACH_FORM,
					XmNleftOffset,       10,
					NULL);
  
    XtAddCallback (okbutton, XmNactivateCallback, (XtCallbackProc) okCB, 
		   (XtPointer) shell);
    XmStringFree (string);


    string = XmStringCreateSimple ("Cancel");
    cancelbutton = XtVaCreateManagedWidget ("setupclose", xmPushButtonWidgetClass, form,
					    XmNlabelString,      string,
					    XmNtopAttachment,    lastform ? XmATTACH_WIDGET : XmATTACH_FORM,
					    XmNtopWidget,        lastform,
					    XmNtopOffset,        10,
					    XmNleftAttachment,   XmATTACH_WIDGET,
					    XmNleftWidget,       okbutton,
					    XmNleftOffset,       10,
					    XmNbottomAttachment, XmATTACH_FORM,
					    XmNbottomOffset,     10,
					    NULL);
  
    XtAddCallback (cancelbutton, XmNactivateCallback, (XtCallbackProc) cancelCB,
		   (XtPointer) shell);
    XmStringFree (string);


    /* manage and map shell widget */

    XtManageChild (shell);
    XtMapWidget (shell);
}





void save_options (void)
{
    FILE *f;
    char filename[256];
    char tmp[256];
    char *ptr;
    int i;
    
    if ((ptr = getenv ("HOME")) != NULL) {
	strcpy (filename, ptr);
    } else {
	strcpy (filename, ".");
    }
    
    strcat (filename, "/.icaorc");
    
    f = fopen (filename, "w");
    
    if (f) {
	fprintf (f, "/* ICAO Map Setup */\n");
	fprintf (f, "\n");
	fprintf (f, "set scale 1:%ld\n", map_scale);
	fprintf (f, "set origin %s\n\n", internal2string (map_origin));
	
	if (map_projection == 0) {
	    fprintf (f, "set projection lambert\n");
	} else { 
	    fprintf (f, "set projection mercator\n");
	}
	
	fprintf (f, "set unit distance ");
	switch (unit_distance) {
	  case km: fprintf (f, "km\n"); break;
	  case nm: fprintf (f, "nm\n"); break;
	}
	
	fprintf (f, "set unit speed ");
	switch (unit_speed) {
	  case kmh: fprintf (f, "kmh\n"); break;
	  case kts: fprintf (f, "knots\n"); break;
	  case mps: fprintf (f, "mps\n"); break;
	}
	
	fprintf (f, "set unit altitude ");
	switch (unit_altitude) {
	  case mtr: fprintf (f, "m\n"); break;
	  case ft: fprintf (f, "ft\n"); break;
	}
	
	/* visible object string */
	
	for (i=0; i<numvis; i++) {
	    tmp[i] = visobj[i] ? '1' : '0';
	}
	tmp[numvis] = 0;
	fprintf (f, "set visible \"%s\"\n", tmp);

	/* distance function */

	fprintf (f, "set algorithm distance ");
	switch (config_distance_func) {
	  case pauly:
	    fprintf (f, "Pauly\n");
	    break;
	  case FAI:
	    fprintf (f, "FAI\n");
	    break;
	}
	
	/* save world file references */

	fprintf (f, "\n/* load world files */\n\n");

	for (i=0; i<next_mapfile; i++) {
	    fprintf (f, "load map %s\n", mapfiles[i]);
	}
	
	fclose (f);
    } else {
	fprintf (stderr, "Unable to write $s\n", filename);
    }
}



/* process visible string from config file */

void set_visiblestring (char *string)
{
    int i;
    int max = strlen (string);
    if (numvis < max) {
	max = numvis;
    }

    for (i=0; i<max; i++) {
	switch (string[i]) {
	  case '0':
	    visobj[i] = 0;
	    break;
	  case '1':
	    visobj[i] = 1;
	    break;
	}
    }
}
